// Written by Marcel Popescu, marcel@aiurea.com
// Placed in the public domain

unit MainFormU;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, FileCtrl, IniFiles;

type
  TMainForm = class(TForm)
    DriveList: TDriveComboBox;
    DirList: TDirectoryListBox;
    FileList: TFileListBox;
    SelList: TListBox;
    Label1: TLabel;
    DoRandom: TButton;
    Label2: TLabel;
    DoXOR: TButton;
    SaveDlg: TSaveDialog;
    DoAbout: TButton;
    DoRandomize: TButton;
    ckShowAll: TCheckBox;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FileListDblClick(Sender: TObject);
    procedure SelListDblClick(Sender: TObject);
    procedure DoRandomClick(Sender: TObject);
    procedure DoXORClick(Sender: TObject);
    procedure DoAboutClick(Sender: TObject);
    procedure DoRandomizeClick(Sender: TObject);
    procedure ckShowAllClick(Sender: TObject);
  private
    { Private declarations }
    FSettings   : TIniFile;
  public
    { Public declarations }
  end;

const
  PADSize       = 128 * 1024 div 4;

type
  PPadBuffer    = ^TPadBuffer;
  TPadBuffer    = array [1 .. PADSize] of Cardinal;

var
  MainForm: TMainForm;

  function      GetPadName (pPad: PPadBuffer): string;
  procedure     XORList (pInput: TStrings; pPad: PPadBuffer);
  procedure     WritePad (pPad: PPadBuffer; const pOutput: string);

implementation

{$R *.DFM}

uses
  General, CommonU, HashU, CipherU, RandomU, SHA_1U, BlowfishU,
  RandomFormU, AboutU;

var
  rnd           : IBetterRNG;

  function      GetPadName;
  var
    s           : string;

  begin
    s := StringOfChar (' ', 8);
    Move (pPad^ [1], s [1], Length (s));

    Result := 'pad' + BufToHex (s) + '.dat';
  end;

  procedure     XORList;
  var
    iStr        : TStream;
    i, j        : Integer;
    r           : Cardinal;

  begin
    // fill the output stream with zeros
    FillChar (pPad^, SizeOf (pPad^), 0);

    // now xor all input streams
    with pInput do
      for i := 0 to Pred (Count) do
        begin
          iStr := TFileStream.Create (Strings [i], fmOpenRead or fmShareDenyWrite);
          try
            for j := Low (pPad^) to High (pPad^) do
              begin
                iStr.Read (r, SizeOf (r));
                pPad^ [j] := pPad^ [j] xor r;
              end;

          finally
            iStr.Free;
          end;
        end;
  end;

  procedure     WritePad;
  var
    f           : file of Cardinal;
    i           : Integer;

  begin
    AssignFile (f, pOutput);
    Rewrite (f);
    try
      for i := Low (pPad^) to High (pPad^) do
        Write (f, pPad^ [i]);

    finally
      CloseFile (f);
    end;
  end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  FSettings := TIniFile.Create (ExtractFilePath (ParamStr (0)) + 'XORFiles.ini');
  with FSettings do
    begin
      DriveList.Drive := ReadString ('global', 'drive', 'C:') [1];
      DirList.Directory := ReadString ('global', 'path', '\');
      ckShowAll.Checked := ReadBool ('global', 'showAll', False);
    end;
  FSettings.Free;

  rnd := TBetterRNG.Create;
  rnd.Hash := GetCrypt (2) as ICryptHash;
  rnd.Cipher := GetCrypt (5) as ICryptCipher;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  rnd := nil;

  FSettings := TIniFile.Create (ExtractFilePath (ParamStr (0)) + 'XORFiles.ini');
  with FSettings do
    begin
      WriteString ('global', 'drive', DriveList.Drive);
      WriteString ('global', 'path', DirList.Directory);
      WriteBool ('global', 'showAll', ckShowAll.Checked);
    end;
  FSettings.Free;
end;

procedure TMainForm.FileListDblClick(Sender: TObject);
var
  s: string;
  f: file;
begin
  s := FileList.FileName;
  if SelList.Items.IndexOf (s) < 0 then
    begin
      AssignFile (f, s); {$I-} Reset (f); {$I+}
      if IOResult <> 0 then
        begin
          ShowMessage (Format ('Error trying to open %s.', [s]));

          Exit;
        end;

      if FileSize (f) <= PADSize * 4 then
        SelList.Items.Add (s)
      else
        begin
          ShowMessage ('Cannot use a file larger than 128K. If you need this, ' +
              'please contact the author at marcel@aiurea.com');

          // ...
        end;

      {$I-} CloseFile (f); {$I+}
      if IOResult <> 0 then ;
    end;
end;

procedure TMainForm.SelListDblClick(Sender: TObject);
begin
  with SelList do
    if ItemIndex < Items.Count then
      Items.Delete (ItemIndex)
end;

procedure TMainForm.DoRandomClick(Sender: TObject);
var
  buf: TPadBuffer;
  i: Integer;
  str: TStream;
begin
  Screen.Cursor := crHourGlass;

  // generate the pad
  for i := Low (buf) to High (buf) do
    buf [i] := rnd.RandCard;

  str := TFileStream.Create (DirList.Directory + '\' + GetPadName (@buf),
      fmCreate or fmShareExclusive);
  try
    str.Write (buf, SizeOf (buf));

  finally
    str.Free;
  end;

  FileList.Update;
  Screen.Cursor := crArrow;
end;

procedure TMainForm.DoXORClick(Sender: TObject);
var
  pad: TPadBuffer;
begin
  Screen.Cursor := crHourGlass;

  if SelList.Items.Count <= 1 then
    begin
      ShowMessage ('I need at least two files.');
      Screen.Cursor := crArrow;

      Exit;
    end;

  Screen.Cursor := crHourGlass;
  XORList (SelList.Items, @pad);
  Screen.Cursor := crArrow;

  SaveDlg.FileName := GetPadName (@pad);
  if SaveDlg.Execute then
    begin
      WritePad (@pad, SaveDlg.FileName);
      FileList.Update;
    end;

  Screen.Cursor := crArrow;
end;

procedure TMainForm.DoAboutClick(Sender: TObject);
begin
  Application.CreateForm (TAboutBox, AboutBox);
  AboutBox.ShowModal;
  AboutBox.Free;
end;

procedure TMainForm.DoRandomizeClick(Sender: TObject);
begin
  Application.CreateForm (TRandomForm, RandomForm);
  with RandomForm do
    try
      ShowModal;
      rnd.Feed (Seed, 96);

    finally
      Free;
    end;
end;

procedure TMainForm.ckShowAllClick(Sender: TObject);
begin
  if ckShowAll.Checked then
    FileList.Mask := '*.*'
  else
    FileList.Mask := 'pad*.dat'
end;

end.

