{$A+,O+,R-}
unit RandomU;

interface

uses
  HashU, CipherU;

type
  IRNG          = interface (IUnknown)
                ['{E20DD4C0-EB73-11D3-8BEE-00C0DFECF5B7}']
    procedure   Reset (const pSeed: string = '');

    function    GetRand (pBytes: Integer): string;
    function    GetRandCard: Cardinal;
    function    GetRandWord: Word;
    function    GetRandByte: Byte;

    property    RandCard: Cardinal read GetRandCard;
    property    RandWord: Word read GetRandWord;
    property    RandByte: Byte read GetRandByte;
  end;

type
  TRNG          = class (TInterfacedObject, IRNG)
  protected
    procedure   Reset (const pSeed: string = ''); virtual; abstract;

    function    GetRand (pBytes: Integer): string; virtual; abstract;
    function    GetRandCard: Cardinal; virtual;
    function    GetRandWord: Word; virtual;
    function    GetRandByte: Byte; virtual;

    property    RandCard: Cardinal read GetRandCard;
    property    RandWord: Word read GetRandWord;
    property    RandByte: Byte read GetRandByte;

  public
    constructor Create; virtual;
  end;

type
  ISimpleRNG    = interface (IRNG)
                ['{A9C1CA20-EB7F-11D3-8BEE-00C0DFECF5B7}']
    function    GetHash: ICryptHash;
    procedure   SetHash (pHash: ICryptHash);

    property    Hash: ICryptHash read GetHash write SetHash;
  end;

type
  TSimpleRNG    = class (TRNG, IRNG, ISimpleRNG)
  protected
    FCounter    : Integer;
    FBuffer     : string;
    FBufPos     : Integer;
    FHash       : ICryptHash;

    procedure   NextBatch; virtual;

    procedure   Reset (const pSeed: string = ''); override;
    function    GetRand (pBytes: Integer): string; override;

    function    GetHash: ICryptHash; virtual;
    procedure   SetHash (pHash: ICryptHash); virtual;

    property    Hash: ICryptHash read GetHash write SetHash;

  public
    constructor Create; override;
    destructor  Destroy; override;
  end;

type
  IRNGPool      = interface;

  TReseedProc   = procedure (pPool: IRNGPool) of object;

  IRNGPool      = interface (IUnknown)
                ['{5965D9E0-EBED-11D3-8BEE-00C0DFECF5B7}']
    procedure   Init (pSpeed: Single; pThreshold: Integer; pRProc: TReseedProc;
                    pHash: ICryptHash);
    procedure   Clear;
    procedure   Feed (const pInput: string; pEstimate: Integer = 0);

    function    EvalEntropy (const pInput: string; pEstimate: Integer = 0): Integer;

    function    GetBuffer: string;
    function    GetEBits: Integer;

    property    Buffer: string read GetBuffer;
    property    EBits: Integer read GetEBits;
  end;

type
  TRNGPool      = class (TInterfacedObject, IRNGPool)
  protected
    FSpeed      : Single;
    FThreshold  : Integer;
    FRProc      : TReseedProc;
    FHash       : ICryptHash;
    FBuffer     : string;
    FEBits      : Integer;

    IBuff,
    OBuff       : string;

    procedure   DoReseed; virtual;

    procedure   Init (pSpeed: Single; pThreshold: Integer; pRProc: TReseedProc;
                    pHash: ICryptHash); virtual;
    procedure   Clear; virtual;
    procedure   Feed (const pInput: string; pEstimate: Integer = 0); virtual;

                // helper routines for EvalEntropy
    procedure   GetBytes (var pData; pSize: Word; var pActualSize: Word);
    procedure   PutBytes (var pData; pSize: Word; var pActualSize: Word);
    function    EvalEntropy (const pInput: string; pEstimate: Integer = 0): Integer; virtual;

    function    GetBuffer: string; virtual;
    function    GetEBits: Integer; virtual;

    property    Buffer: string read GetBuffer;
    property    EBits: Integer read GetEBits;
  end;

type
  IBetterRNG    = interface (ISimpleRNG)
                ['{E20DD4C1-EB73-11D3-8BEE-00C0DFECF5B7}']
    function    GetCipher: ICryptCipher;
    procedure   SetCipher (pCipher: ICryptCipher);

    procedure   Feed (const pEntropy: string; pEstimate: Integer = 0);
    procedure   Reseed (pPool: IRNGPool);

    property    Cipher: ICryptCipher read GetCipher write SetCipher;
  end;

type
  TBetterRNG    = class (TSimpleRNG, IRNG, ISimpleRNG, IBetterRNG)
  protected
    FFastPool   : IRNGPool;
    FSlowPool   : IRNGPool;
    FLastPool   : IRNGPool;
    FCounter    : Cardinal;
    FKey        : string;
    FCipher     : ICryptCipher;

    procedure   SetHash (pHash: ICryptHash); override;

    procedure   NextBatch; override;
    procedure   Reset (const pSeed: string = ''); override;

    function    GetCipher: ICryptCipher; virtual;
    procedure   SetCipher (pCipher: ICryptCipher); virtual;

    procedure   Feed (const pInput: string; pEstimate: Integer = 0); virtual;
    procedure   Reseed (pPool: IRNGPool); virtual;

    property    Cipher: ICryptCipher read GetCipher write SetCipher;

  public
    constructor Create; override;
    destructor  Destroy; override;
  end;

implementation

uses
  SysUtils, Math,
  LZH;

const
  KEY_THR       = 512;
  ENT_FAST_THR  = 100;
  ENT_SLOW_THR  = 160;

// TRNG

  function      TRNG.GetRandCard;
  var
    s           : string;
    r           : Cardinal;

  begin
    s := GetRand (SizeOf (r));
    Move (s [1], r, SizeOf (r));

    Result := r;
  end;

  function      TRNG.GetRandWord;
  var
    s           : string;
    r           : Word;

  begin
    s := GetRand (SizeOf (r));
    Move (s [1], r, SizeOf (r));

    Result := r;
  end;

  function      TRNG.GetRandByte;
  var
    s           : string;
    r           : Byte;

  begin
    s := GetRand (SizeOf (r));
    Move (s [1], r, SizeOf (r));

    Result := r;
  end;

  constructor   TRNG.Create;
  begin
    inherited Create;
  end;

// TSimpleRNG

  procedure     TSimpleRNG.NextBatch;
  begin
    Inc (FCounter);
    Reset (FBuffer + IntToStr (FCounter));
  end;

  procedure     TSimpleRNG.Reset;
  begin
    FCounter := 0;
    FBuffer := HashStr (FHash, pSeed);
    FBufPos := 1;
  end;

  function      TSimpleRNG.GetRand;
  var
    i           : Integer;

  begin
    Result := '';

    for i := 1 to pBytes do
      begin
        if FBufPos > Length (FBuffer) then
          NextBatch;

        Result := Result + FBuffer [FBufPos];
        Inc (FBufPos);
      end;
  end;

  function      TSimpleRNG.GetHash;
  begin
    Result := FHash
  end;

  procedure     TSimpleRNG.SetHash;
  begin
    FHash := pHash;

    if Assigned (FHash) then
      Reset ('');
  end;

  constructor   TSimpleRNG.Create;
  begin
    inherited Create;
  end;

  destructor    TSimpleRNG.Destroy;
  begin
    inherited Destroy;
  end;

// TRNGPool

  procedure     TRNGPool.DoReseed;
  begin
    if Assigned (FRProc) then
      FRProc (Self as IRNGPool);

    Clear;
  end;

  procedure     TRNGPool.Init;
  begin
    FSpeed := pSpeed;
    FThreshold := pThreshold;
    FRProc := pRProc;
    FHash := pHash;

    Clear;
  end;

  procedure     TRNGPool.Clear;
  begin
    FBuffer := '';
    FEBits := 0;
  end;

  procedure     TRNGPool.Feed;
  begin
    Inc (FEBits, EvalEntropy (pInput, pEstimate));

    FBuffer := FBuffer + HashStr (FHash, pInput);
    if FEBits >= FThreshold then
      DoReseed;
  end;

  procedure     TRNGPool.GetBytes;
  begin
    pActualSize := Min (pSize, Length (IBuff));
    Move (IBuff [1], pData, pActualSize);
    Delete (IBuff, 1, pActualSize);
  end;

  procedure     TRNGPool.PutBytes;
  var
    s           : string;

  begin
    pActualSize := pSize;
    SetLength (s, pActualSize);
    Move (pData, s [1], pActualSize);
    OBuff := OBuff + s;
  end;

  function      TRNGPool.EvalEntropy;
  var
    lzh         : TLZH;
    temp        : Integer;

  begin
    if pInput = '' then
      begin
        Result := 0;
        Exit;
      end;

    lzh := TLZH.Create;
    with lzh do
      try
        IBuff := pInput;
        LZHPack (temp, GetBytes, PutBytes);

        // the estimated entropy must be returned in BITS
        Result := Length (OBuff) * 8;

        // now, the result must not be bigger than Length (pInput) * 8 * FSpeed
        Result := Min (Result, Trunc (Length (pInput) * 8 * FSpeed));

        // finally, if the user had an estimate, that's the maximum
        if pEstimate > 0 then
          Result := Min (Result, pEstimate);

      finally
        Free;
      end;
  end;

  function      TRNGPool.GetBuffer;
  begin
    Result := HashStr (FHash, FBuffer)
  end;

  function      TRNGPool.GetEBits;
  begin
    Result := FEBits
  end;

// TBetterRNG

  procedure     TBetterRNG.SetHash;
  begin
    FHash := pHash;

    if Assigned (FHash) then
      begin
        FFastPool.Init (0.5, ENT_FAST_THR, Reseed, FHash);
        FSlowPool.Init (0.3, ENT_SLOW_THR, Reseed, FHash);
      end;
  end;

  procedure     TBetterRNG.NextBatch;
  var
    s           : string;

  begin
    Inc (FCounter);

    SetLength (s, SizeOf (FCounter));
    Move (FCounter, s [1], Length (s));

    FBuffer := CipherStr (FHash, FCipher, FKey, s);
    FBufPos := 1;

    // refresh the key?
    if FCounter mod KEY_THR = 0 then
      begin
        SetLength (FKey, FCipher.GetMaxKeySize div 8);
        FKey := GetRand (Length (FKey));
      end;
  end;

  procedure     TBetterRNG.Reset;
  begin
    FFastPool.Clear;
    FSlowPool.Clear;
    FLastPool := FFastPool;

    FCounter := 1;
    FKey := HashStr (FHash, pSeed);
    FBuffer := '';
    FBufPos := 1;
  end;

  function      TBetterRNG.GetCipher;
  begin
    Result := FCipher
  end;

  procedure     TBetterRNG.SetCipher;
  begin
    FCipher := pCipher;

    if Assigned (FCipher) then
      Reset ('');
  end;

  procedure     TBetterRNG.Feed;
  begin
    FLastPool.Feed (pInput, pEstimate);

    if FLastPool = FFastPool then
      FLastPool := FSlowPool
    else
      FLastPool := FFastPool;
  end;

  procedure     TBetterRNG.Reseed;
  var
    s           : string;

  begin
    s := FKey + pPool.Buffer;
    if pPool = FSlowPool then
      begin
        s := s + FSlowPool.Buffer;
        FSlowPool.Clear;
      end;

    FKey := Copy (HashStr (FHash, s), 1, (FCipher.GetMaxKeySize div 8))
  end;

  constructor   TBetterRNG.Create;
  begin
    inherited Create;

    FFastPool := TRNGPool.Create as IRNGPool;
    FSlowPool := TRNGPool.Create as IRNGPool;
  end;

  destructor    TBetterRNG.Destroy;
  begin
    FSlowPool := nil;
    FFastPool := nil;

    inherited Destroy;
  end;

end.

